"""Client implementation."""


import json
import os
from concurrent.futures import ThreadPoolExecutor

import requests

from .utils import _prepare_input, _prepare_output


class Client:
    """
    Client is a representation of a client.

    :param address: address of a server like http://localhost:3000
    """

    def __init__(self, address):
        self._address = os.path.join(address, 'jsonrpc')
        self._id = 0
        self._pool = ThreadPoolExecutor(1)

    def _increment_id(self):
        self._id += 1
        if self._id > 10 ** 8:
            self._id = 0

    def _send_request(self, method, data):
        payload = {
            'method': method,
            'params': [_prepare_input(data)],
            'jsonrpc': '2.0',
            'id': self._id,
        }

        self._increment_id()

        response = requests.post(
            self._address,
            data=json.dumps(payload),
            headers={'content-type': 'application/json'}
        ).json()

        if response.get('error', None) is not None:
            raise ValueError(response['error'])

        return _prepare_output(response['result'])

    def _ping(self):
        payload = {
            'method': 'ready',
            'params': [],
            'jsonrpc': '2.0',
            'id': self._id,
        }

        self._increment_id()

        response = requests.post(
            self._address,
            data=json.dumps(payload),
            headers={'content-type': 'application/json'}
        ).json()

        return response['result']

    def ready(self):
        """
        ready tells if server is ready.

        :returns: true or false
        """
        try:
            return self._ping()
        except requests.exceptions.ConnectionError:
            return False

    def started(self):
        """
        sstarted tells if server started.

        :returns: true or false
        """
        try:
            _ = self._ping()
            return True
        except requests.exceptions.ConnectionError:
            return False

    def train(self, data):
        """
        train launches training ml algorithm on server on provided data.

        :param data: data to process
        """
        return self._pool.submit(self._send_request, 'train', data)

    def update(self, data):
        """
        update sends new data to state machine.

        :param data: data to process
        """

        return self._pool.submit(self._send_request, 'update', data)

    def set_state(self, data):
        """
        set_state sends new state data to state machine.

        :param data: data to process
        """

        return self._pool.submit(self._send_request, 'set_state', data)

    def predict(self, data):
        """
        predict launches prediction ml algorithm on server on provided data.

        :param data: data to process
        """

        return self._pool.submit(self._send_request, 'predict', data)



"""StateMachine implementation."""

from .base import BaseServer
from .utils import serialize_data


class StateMachine(BaseServer):
    """
    StateMachine is a representation of a state machine.

    :param port: port in form of integer
    :param log: is logging enabled or not
    :param ml: interface of ml object, need to implement predict method
    :param address: address of a server, keep default value if you don't understand what it's for
    """

    def __init__(self, port, ml, address='0.0.0.0',
                 ml_config=None, log=True):
        if not all(method in dir(ml) for method in [
                'update', 'set_state']):
            raise ValueError('ml must have update and set_state method')

        super().__init__(port, ml, address, ml_config, log)

        self.set_dispatcher({
            'update': self._update,
            'set_state': self._set_state
        })

    @serialize_data
    def _update(self, data):
        return self._ml.update(data)

    @serialize_data
    def _set_state(self, data):
        return self._ml.set_state(data)



"""preparator contains EchoPreparator class."""


class EchoPreparator:
    """EchoPreparator returns input and output as it is."""

    @staticmethod
    def prepare_input(data):
        """prepare_input prepares input."""
        return data

    @staticmethod
    def prepare_output(data):
        """prepare_output prepares output."""
        return data



"""BaseServer implementation."""


import logging
from concurrent.futures import ThreadPoolExecutor
from copy import copy

from jsonrpc import JSONRPCResponseManager, dispatcher
from werkzeug.serving import run_simple
from werkzeug.wrappers import Request, Response


class BaseServer:
    """
    Server is a representation of a server.

    :param port: port in form of integer
    :param ml: interface of ml object, need to implement predict method
    :param log: turn logging off or on. True by default.
    :param address: address of a server, keep default value if you don't understand what it's for
    """

    def __init__(self, port, ml, address='0.0.0.0', ml_config=None, log=True):
        if not log:
            logging.getLogger('werkzeug').setLevel(logging.ERROR)
            for name, _ in logging.Logger.manager.loggerDict.items():
                if any([x in name for x in ['requests', 'werkzeug', 'urllib']]):
                    logging.getLogger(name).setLevel(logging.ERROR)
        self._address = address
        self._port = port
        self._pool = ThreadPoolExecutor(1)
        self._dispatcher = copy(dispatcher)

        self._ml_config = ml_config
        if ml_config is not None:
            self._ml_constructor = ml
            self._ml = None
        else:
            self._ml = ml

        self.set_dispatcher({'ready': self._ready})

    def _init_ml(self):
        self._ml = self._ml_constructor(**self._ml_config)

    def _ready(self):
        return self._ml is not None

    def set_dispatcher(self, route_table, clear=False):
        """
        set_dispatcher sets dispacther keys and values.

        :param route_table: route_table is a dict with keys as routes and values as callables
        """

        if clear:
            self._dispatcher = copy(dispatcher)
        for route, func in route_table.items():
            self._dispatcher[route] = func

        return self

    @staticmethod
    def _init_cb(future):
        if future.exception() is not None:
            raise RuntimeError(future.exception())

    def __call__(self):
        if self._ml_config is not None:
            self._pool.submit(self._init_ml).add_done_callback(self._init_cb)

        @Request.application
        def _application(request):
            # dispatcher is dictionary {<method_name>: callable}
            response = JSONRPCResponseManager.handle(
                request.data, self._dispatcher)
            return Response(response.json, mimetype='application/json')

        run_simple(self._address, self._port, _application)



"""utils for clients"""

import codecs
import pickle


def _prepare_input(data):
    return codecs.encode(pickle.dumps(data), 'base64').decode()


def _prepare_output(data):
    return pickle.loads(codecs.decode(bytes(data, encoding='utf-8'), 'base64'))



"""utils for servers"""

import codecs
import pickle


def serialize_data(func):
    """parse_data is a decorator which serialize incoming and outgoing data."""
    def _function_wrapper(*args):
        return _prepare_output(func(args[0], _prepare_input(args[-1])))

    return _function_wrapper


def _prepare_output(data):
    return codecs.encode(pickle.dumps(data), 'base64').decode()


def _prepare_input(data):
    return pickle.loads(codecs.decode(bytes(data, encoding='utf-8'), 'base64'))



"""Server implementation."""


from .base import BaseServer
from .preparator import EchoPreparator
from .utils import serialize_data


class Server(BaseServer):
    """
    Server is a representation of a server.

    :param port: port in form of integer
    :param log: is logging enabled or not
    :param ml: interface of ml object, need to implement predict method
    :param address: address of a server, keep default value if you don't understand what it's for
    """

    def __init__(self, port, ml, address='0.0.0.0', preparator=EchoPreparator,
                 ml_config=None, log=True):
        if not any(method in dir(ml) for method in [
                'predict', 'train']):
            raise ValueError('ml must have predict or train method')

        if not all(method in dir(preparator) for method in [
                'prepare_output', 'prepare_input']):
            raise ValueError(
                'preparator must have prepare_output and prepare_input methods')

        super().__init__(port, ml, address, ml_config, log)

        self.set_dispatcher({
            'train': self._train,
            'predict': self._predict
        })

        self._preparator = preparator

    @serialize_data
    def _predict(self, data):
        return self._ml.predict(data)

    @serialize_data
    def _train(self, data):
        prepared_data = self._preparator.prepare_input(data)
        res = self._ml.train(prepared_data)
        prepared_res = self._preparator.prepare_output(res)
        return prepared_res



# pylint: skip-file

import distutils.cmd
import distutils.log
import os
import subprocess

from setuptools import find_packages, setup

with open('requirements.txt') as f:
    required = f.read().splitlines()


class PylintCommand(distutils.cmd.Command):
    """A custom command to run Pylint on all Python source files."""

    description = 'run Pylint on Python source files'
    user_options = [
        # The format is (long option, short option, description).
        ('pylint-rcfile=', None, 'path to Pylint config file'),
    ]

    def initialize_options(self):
        """Set default values for options."""
        # Each user option must be listed here with their default value.
        self.pylint_rcfile = '.pylintrc'

    def finalize_options(self):
        """Post-process options."""
        if self.pylint_rcfile:
            assert os.path.exists(self.pylint_rcfile), (
                'Pylint config file %s does not exist.' % self.pylint_rcfile)

    def run(self):
        """Run command."""
        command = ['pylint']
        if self.pylint_rcfile:
            command.append('--rcfile=%s' % self.pylint_rcfile)
        command.append(os.getcwd() + '/mls')
        self.announce(
            'Running command: %s' % str(command),
            level=distutils.log.INFO)
        subprocess.check_call(command)


tests_require = [
    'numpy>=1.12.1',
]

CONFIG = {
    'name': 'myLittleServer',
    'url': 'https://github.com/connectome-ai/mls',
    'download_url': 'https://github.com/connectome-ai/mls/archive/1.1.tar.gz',
    'version': '1.1',
    'description': 'mls is a wrapper around ml code.',
    'author': 'connectome.ai',
    'test_suite': 'mls',
    'packages': find_packages(exclude=['tests', '*.tests', '*.tests.*']),
    'tests_require': tests_require,
    'install_requires': required,
    'cmdclass': {
        'pylint': PylintCommand
    },
}

setup(**CONFIG)
